@@ -1,14 +1,21 @@
 /*
-   Copyright (c) 2003, 2014, Oracle and/or its affiliates. All rights reserved.
+   Copyright (c) 2003, 2017, Oracle and/or its affiliates. All rights reserved.
 
    This program is free software; you can redistribute it and/or modify
-   it under the terms of the GNU General Public License as published by
-   the Free Software Foundation; version 2 of the License.
+   it under the terms of the GNU General Public License, version 2.0,
+   as published by the Free Software Foundation.
+
+   This program is also distributed with certain software (including
+   but not limited to OpenSSL) that is licensed under separate terms,
+   as designated in a particular file or component or in included license
+   documentation.  The authors of MySQL hereby grant you an additional
+   permission to link the program and your derivative works with the
+   separately licensed software that they have included with MySQL.
 
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-   GNU General Public License for more details.
+   GNU General Public License, version 2.0, for more details.
 
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
@@ -139,7 +146,12 @@ public:
                                       in clusterlog */
     CmvmiTestLongSig = 2605,  /* Long signal testing trigger */
     DumpEventLog = 2606,
+
+    CmvmiLongSignalMemorySnapshotStart = 2607,
+    CmvmiLongSignalMemorySnapshot = 2608,
+    CmvmiLongSignalMemorySnapshotCheck = 2609,
     CmvmiSetKillerWatchdog = 2610,
+
     LCPContinue = 5900,
     // 7000 DIH
     // 7001 DIH
@@ -180,6 +192,7 @@ public:
      * Allows GCP stop thresholds to be set
      */
     DihSetGcpStopVals = 7026,
+    DihStallLcpStart = 7027,
     DihDumpPageRecInfo = 7032,
     EnableUndoDelayDataWrite = 7080, // DIH+ACC+TUP
     DihSetTimeBetweenGcp = 7090,
@@ -190,8 +203,13 @@ public:
     TuxSetLogFlags = 12002,
     TuxMetaDataJunk = 12009,
     
-    DumpTsman = 9800, 
+    DumpTsman = 9800,
+ 
     DumpLgman = 10000,
+    LgmanDumpUndoStateClusterLog = 10001,
+    LgmanDumpUndoStateLocalLog = 10002,
+    LgmanCheckCallbacksClear = 10003,
+    
     DumpPgman = 11000,
     DumpBackup = 13000,
     DumpBackupSetCompressed = 13001,
