@@ -1,14 +1,21 @@
 /*
-   Copyright (c) 2003, 2010, Oracle and/or its affiliates. All rights reserved.
+   Copyright (c) 2003, 2015, Oracle and/or its affiliates. All rights reserved.
 
    This program is free software; you can redistribute it and/or modify
-   it under the terms of the GNU General Public License as published by
-   the Free Software Foundation; version 2 of the License.
+   it under the terms of the GNU General Public License, version 2.0,
+   as published by the Free Software Foundation.
+
+   This program is also distributed with certain software (including
+   but not limited to OpenSSL) that is licensed under separate terms,
+   as designated in a particular file or component or in included license
+   documentation.  The authors of MySQL hereby grant you an additional
+   permission to link the program and your derivative works with the
+   separately licensed software that they have included with MySQL.
 
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-   GNU General Public License for more details.
+   GNU General Public License, version 2.0, for more details.
 
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
@@ -36,15 +43,18 @@
 #include <sys/lwp.h>
 #include <sys/processor.h>
 #include <sys/procset.h>
+#include <sys/pset.h>
 #endif
 
 #ifdef NDB_MUTEX_DEADLOCK_DETECTOR
 #include "NdbMutex_DeadlockDetector.h"
 #endif
 
+#if defined(HAVE_LINUX_SCHEDULING) || defined(HAVE_PTHREAD_SET_SCHEDPARAM)
 static int g_min_prio = 0;
 static int g_max_prio = 0;
-static int g_prio = 0;
+static my_bool get_prio_first = TRUE;
+#endif
 
 static NdbMutex *g_ndb_thread_mutex = 0;
 static struct NdbCondition * g_ndb_thread_condition = 0;
@@ -61,14 +71,43 @@ struct NdbThread
 { 
   volatile int inited;
   pthread_t thread;
-#if defined HAVE_SOLARIS_AFFINITY
-  id_t tid;
-#elif defined HAVE_LINUX_SCHEDULING
-  pid_t tid;
+#ifdef _WIN32
+  /*
+    Problem in mysys on certain MySQL versions where the thread id is
+    used as the thread identifier. Since the thread id may be reused
+    by another thread(when the current thread exits) one may end up
+    waiting for the wrong thread.
+
+    Workaround by using a HANDLE which is opened by thread itself in
+    'ndb_thread_wrapper' and subsequently used to wait for the
+    thread in 'NdbThread_WaitFor'.
+
+    NOTE: Windows implementation of 'pthread_join' and support for non
+    detached threads as implemented in MySQL Cluster 7.0 and 7.1 is not
+    affected, only MySQL Cluster based on 5.5+ affected(slightly
+    different implementation). The current workaround in NdbThread is
+    generic and works regardless of pthread_join implementation in mysys.
+  */
+  HANDLE thread_handle;
 #endif
+#if defined HAVE_LINUX_SCHEDULING
+  pid_t tid;
+  struct NdbCpuSet *orig_cpu_set;
+#elif defined HAVE_SOLARIS_AFFINITY
+  /* Our thread id */
+  id_t tid;
+  /* Have we called any lock to CPU function yet for this thread */
+  my_bool first_lock_call;
+  /* Original processor set locked to */
+  psetid_t orig_proc_set;
+  /* Original processor locked to */
+  processorid_t orig_processor_id;
+#endif
+  const struct processor_set_handler *cpu_set_key;
   char thread_name[16];
   NDB_THREAD_FUNC * func;
   void * object;
+  void *thread_key;
 #ifdef NDB_MUTEX_DEADLOCK_DETECTOR
   struct ndb_mutex_thr_state m_mutex_thr_state;
 #endif
@@ -95,9 +134,7 @@ static
 void
 settid(struct NdbThread * thr)
 {
-#if defined HAVE_SOLARIS_AFFINITY
-  thr->tid = _lwp_self();
-#elif defined HAVE_LINUX_SCHEDULING
+#if defined HAVE_LINUX_SCHEDULING
   thr->tid = syscall(SYS_gettid);
   if (thr->tid == (pid_t)-1)
   {
@@ -108,18 +145,24 @@ settid(struct NdbThread * thr)
     */
     thr->tid = getpid();
   }
+#elif defined HAVE_SOLARIS_AFFINITY
+  thr->tid = (id_t)_lwp_self();
+#else
+  (void)thr;
 #endif
 }
 
 int
 NdbThread_GetTid(struct NdbThread* thr)
 {
-#if defined HAVE_SOLARIS_AFFINITY
+#if defined HAVE_LINUX_SCHEDULING
   return (int)thr->tid;
-#elif defined HAVE_LINUX_SCHEDULING
+#elif defined HAVE_SOLARIS_AFFINITY
   return (int)thr->tid;
-#endif
+#else
+  (void)thr;
   return -1;
+#endif
 }
 
 static
@@ -149,10 +192,19 @@ ndb_thread_wrapper(void* _ss){
       struct NdbThread * ss = (struct NdbThread *)_ss;
       settid(ss);
 
+#ifdef _WIN32
+      /*
+        Create the thread handle, ignore failure since it's unlikely
+        to fail and the functions using this handle checks for NULL.
+      */
+      ss->thread_handle =
+        OpenThread(SYNCHRONIZE, FALSE, GetCurrentThreadId());
+#endif
+
 #ifdef NDB_MUTEX_DEADLOCK_DETECTOR
       ndb_mutex_thread_init(&ss->m_mutex_thr_state);
 #endif
-
+      NdbThread_SetTlsKey(NDB_THREAD_TLS_NDB_THREAD, ss);
       NdbMutex_Lock(g_ndb_thread_mutex);
       ss->inited = 1;
       NdbCondition_Signal(g_ndb_thread_condition);
@@ -172,7 +224,7 @@ struct NdbThread*
 NdbThread_CreateObject(const char * name)
 {
   struct NdbThread* tmpThread;
-  DBUG_ENTER("NdbThread_Create");
+  DBUG_ENTER("NdbThread_CreateObject");
 
   if (g_main_thread != 0)
   {
@@ -269,6 +321,12 @@ NdbThread_Create(NDB_THREAD_FUNC *p_thread_func,
   tmpThread->inited = 0;
   tmpThread->func= p_thread_func;
   tmpThread->object= p_thread_arg;
+  tmpThread->cpu_set_key = NULL;
+#ifdef HAVE_LINUX_SCHEDULING
+  tmpThread->orig_cpu_set = NULL;
+#elif HAVE_SOLARIS_AFFINITY
+  tmpThread->first_lock_call = TRUE;
+#endif
 
   NdbMutex_Lock(g_ndb_thread_mutex);
   result = pthread_create(&tmpThread->thread,
@@ -280,7 +338,7 @@ NdbThread_Create(NDB_THREAD_FUNC *p_thread_func,
 
   if (result != 0)
   {
-    NdbMem_Free((char *)tmpThread);
+    NdbMem_Free(tmpThread);
     NdbMutex_Unlock(g_ndb_thread_mutex);
     DBUG_RETURN(0);
   }
@@ -312,7 +370,19 @@ void NdbThread_Destroy(struct NdbThread** p_thread)
 {
   DBUG_ENTER("NdbThread_Destroy");
   if (*p_thread != NULL){
+#ifdef _WIN32
+    HANDLE thread_handle = (*p_thread)->thread_handle;
+    if (thread_handle)
+      CloseHandle(thread_handle);
+#endif
     DBUG_PRINT("enter",("*p_thread: 0x%lx", (long) *p_thread));
+#ifdef HAVE_LINUX_SCHEDULING
+    if ((*p_thread)->orig_cpu_set)
+    {
+      free((*p_thread)->orig_cpu_set);
+      (*p_thread)->orig_cpu_set = NULL;
+    }
+#endif
     free(* p_thread); 
     * p_thread = 0;
   }
@@ -322,17 +392,35 @@ void NdbThread_Destroy(struct NdbThread** p_thread)
 
 int NdbThread_WaitFor(struct NdbThread* p_wait_thread, void** status)
 {
-  int result;
-
   if (p_wait_thread == NULL)
     return 0;
 
   if (p_wait_thread->thread == 0)
     return 0;
 
-  result = pthread_join(p_wait_thread->thread, status);
-  
-  return result;
+#ifdef _WIN32
+  {
+    DWORD ret;
+    HANDLE thread_handle = p_wait_thread->thread_handle;
+
+    if (thread_handle == NULL)
+    {
+      return -1;
+    }
+
+    ret = WaitForSingleObject(thread_handle, INFINITE);
+    if (ret != WAIT_OBJECT_0)
+    {
+      return -1;
+    }
+
+    /* Don't fill in "status", never used anyway */
+
+    return 0;
+  }
+#else
+  return pthread_join(p_wait_thread->thread, status);
+#endif
 }
 
 
@@ -353,6 +441,7 @@ int NdbThread_SetConcurrencyLevel(int level)
 #endif
 }
 
+#if defined(HAVE_LINUX_SCHEDULING) || defined(HAVE_PTHREAD_SET_SCHEDPARAM)
 static int
 get_max_prio(int policy)
 {
@@ -383,14 +472,16 @@ get_min_prio(int policy)
 }
 
 static int
-get_prio(my_bool rt_prio, my_bool high_prio, int policy)
+get_prio(my_bool high_prio, int policy)
 {
-  if (!rt_prio)
-    return 0;
-  if (g_prio != 0)
-    return g_prio;
-  g_max_prio = get_max_prio(policy);
-  g_min_prio = get_min_prio(policy);
+  int prio;
+
+  if (get_prio_first)
+  {
+    g_max_prio = get_max_prio(policy);
+    g_min_prio = get_min_prio(policy);
+    get_prio_first = FALSE;
+  }
   /*
     We need to distinguish between high and low priority threads. High
     priority threads are the threads that don't execute the main thread.
@@ -401,12 +492,29 @@ get_prio(my_bool rt_prio, my_bool high_prio, int policy)
     extensions where a new priority level is required.
   */
   if (high_prio)
-    g_prio = g_min_prio + 3;
+    prio = g_min_prio + 3;
   else
-    g_prio = g_min_prio + 1;
-  if (g_prio < g_min_prio)
-    g_prio = g_min_prio;
-  return g_prio;
+    prio = g_min_prio + 1;
+  if (prio < g_min_prio)
+    prio = g_min_prio;
+  return prio;
+}
+#endif
+
+/**
+ * When running in real-time mode we stop everyone else from running,
+ * we make a short break to give others a chance to execute that are
+ * on lower prio. If we don't do this regularly and run for a long
+ * time on real-time prio then we can easily crash the system.
+ */
+int
+NdbThread_yield_rt(struct NdbThread* pThread, my_bool high_prio)
+{
+  int res = NdbThread_SetScheduler(pThread, FALSE, high_prio);
+  int res1 = NdbThread_SetScheduler(pThread, TRUE, high_prio);
+  if (res || res1)
+    return res;
+  return 0;
 }
 
 int
@@ -421,7 +529,7 @@ NdbThread_SetScheduler(struct NdbThread* pThread,
   if (rt_prio)
   {
     policy = SCHED_RR;
-    prio = get_prio(rt_prio, high_prio, policy);
+    prio = get_prio(high_prio, policy);
   }
   else
   {
@@ -443,7 +551,7 @@ NdbThread_SetScheduler(struct NdbThread* pThread,
   if (rt_prio)
   {
     policy = SCHED_RR;
-    prio = get_prio(rt_prio, high_prio, policy);
+    prio = get_prio(high_prio, policy);
   }
   else
   {
@@ -456,6 +564,9 @@ NdbThread_SetScheduler(struct NdbThread* pThread,
   if (ret)
     error_no= errno;
 #else
+  (void)rt_prio;
+  (void)high_prio;
+  (void)pThread;
   error_no = ENOSYS;
 #endif
 
@@ -463,9 +574,269 @@ NdbThread_SetScheduler(struct NdbThread* pThread,
 }
 
 int
-NdbThread_LockCPU(struct NdbThread* pThread, Uint32 cpu_id)
+NdbThread_UnlockCPU(struct NdbThread* pThread)
+{
+  int ret;
+  int error_no = 0;
+
+#if defined HAVE_LINUX_SCHEDULING
+ if (pThread->orig_cpu_set != NULL)
+ {
+    cpu_set_t *cpu_set_ptr = (cpu_set_t *)pThread->orig_cpu_set;
+    ret= sched_setaffinity(pThread->tid, sizeof(cpu_set_t), cpu_set_ptr);
+    if (ret)
+    {
+      error_no = errno;
+    }
+    free(pThread->orig_cpu_set);
+    pThread->orig_cpu_set = NULL;
+  }
+#elif defined HAVE_SOLARIS_AFFINITY
+  if (!pThread->first_lock_call)
+  {
+    ret= pset_bind(pThread->orig_proc_set,
+                   P_LWPID,
+                   pThread->tid,
+                   NULL);
+    if (ret)
+    {
+      error_no = errno;
+    }
+    else
+    {
+      ret= processor_bind(P_LWPID,
+                          pThread->tid,
+                          pThread->orig_processor_id,
+                          NULL);
+      if (ret)
+      {
+        error_no = errno;
+      }
+      else
+      {
+        pThread->first_lock_call = TRUE;
+      }
+    }
+  }
+#else
+  (void)ret;
+#endif
+  pThread->cpu_set_key = NULL;
+  return error_no;
+}
+
+static int
+set_old_cpu_locking(struct NdbThread* pThread)
 {
   int error_no = 0;
+  int ret;
+#if defined HAVE_LINUX_SCHEDULING
+  if (pThread->orig_cpu_set == NULL)
+  {
+    cpu_set_t *old_cpu_set_ptr = malloc(sizeof(cpu_set_t));
+    if (!old_cpu_set_ptr)
+    {
+      error_no = errno;
+      goto end;
+    }
+    ret = sched_getaffinity(pThread->tid, sizeof(cpu_set_t), old_cpu_set_ptr);
+    if (ret)
+    {
+      error_no = errno;
+      goto end;
+    }
+    pThread->orig_cpu_set = (struct NdbCpuSet*)old_cpu_set_ptr;
+  }
+#elif defined HAVE_SOLARIS_AFFINITY
+  if (pThread->first_lock_call)
+  {
+    ret = pset_bind(PS_QUERY, P_LWPID, pThread->tid, &pThread->orig_proc_set);
+    if (ret)
+    {
+      error_no= errno;
+      goto end;
+    }
+    ret= processor_bind(P_LWPID,
+                        pThread->tid,
+                        PBIND_QUERY,
+                        &pThread->orig_processor_id);
+    if (ret)
+    {
+      error_no= errno;
+      goto end;
+    }
+    pThread->first_lock_call = FALSE;
+  }
+#else
+  (void)ret;
+  (void)pThread;
+  goto end;
+#endif
+end:
+  return error_no;
+}
+
+int
+NdbThread_LockCreateCPUSet(const Uint32 *cpu_ids,
+                           Uint32 num_cpu_ids,
+                           struct NdbCpuSet **cpu_set)
+{
+#if defined HAVE_LINUX_SCHEDULING
+  int error_no;
+  Uint32 cpu_id;
+  Uint32 i;
+  cpu_set_t *cpu_set_ptr = malloc(sizeof(cpu_set_t));
+
+  if (!cpu_set_ptr)
+  {
+    error_no = errno;
+    *cpu_set = NULL;
+    return error_no;
+  }
+  CPU_ZERO(cpu_set_ptr);
+  for (i = 0; i < num_cpu_ids; i++)
+  {
+    cpu_id = cpu_ids[i];
+    CPU_SET(cpu_id, cpu_set_ptr);
+  }
+  *cpu_set = (struct NdbCpuSet*)cpu_set_ptr;
+  return 0;
+#elif defined HAVE_SOLARIS_AFFINITY
+  int ret;
+  int error_no;
+  Uint32 i;
+  psetid_t *cpu_set_ptr = malloc(sizeof(psetid_t));
+
+  if (!cpu_set_ptr)
+  {
+    error_no = errno;
+    goto end_error;
+  }
+
+  if ((ret = pset_create(cpu_set_ptr)))
+  {
+    error_no = errno;
+    goto error;
+  }
+
+  for (i = 0; i < num_cpu_ids; i++)
+  {
+    if ((ret = pset_assign(*cpu_set_ptr, cpu_ids[i], NULL)))
+    {
+      error_no = errno;
+      goto late_error;
+    }
+  }
+  *cpu_set = (struct NdbCpuSet*)cpu_set_ptr;
+  return 0;
+
+late_error:
+  pset_destroy(*cpu_set_ptr);
+error:
+  free(cpu_set_ptr);
+end_error:
+  *cpu_set = NULL;
+  return error_no;
+
+#else
+  (void)num_cpu_ids;
+  (void)cpu_ids;
+  *cpu_set = NULL;
+  return 0;
+#endif
+}
+
+void
+NdbThread_LockDestroyCPUSet(struct NdbCpuSet *cpu_set)
+{
+  if (cpu_set != NULL)
+  {
+#if defined HAVE_LINUX_SCHEDULING
+  /* Empty */
+#elif defined HAVE_SOLARIS_AFFINITY
+    pset_destroy(*((psetid_t*)cpu_set));
+#endif
+    free(cpu_set);
+  }
+}
+
+int
+NdbThread_LockCPUSet(struct NdbThread* pThread,
+                     struct NdbCpuSet *ndb_cpu_set,
+                     const struct processor_set_handler *cpu_set_key)
+{
+  int error_no;
+  int ret;
+#if defined(HAVE_LINUX_SCHEDULING)
+  cpu_set_t *cpu_set_ptr;
+#elif defined(HAVE_SOLARIS_AFFINITY)
+  psetid_t *cpu_set_ptr;
+#endif
+
+  if ((error_no = set_old_cpu_locking(pThread)))
+    goto end;
+
+  if (ndb_cpu_set == NULL)
+  {
+    return 0;
+  }
+#if defined HAVE_LINUX_SCHEDULING
+  cpu_set_ptr = (cpu_set_t*)ndb_cpu_set;
+
+  /* Lock against the bitmask defined by CPUSet */
+  ret= sched_setaffinity(pThread->tid,
+                         sizeof(cpu_set_t),
+                         cpu_set_ptr);
+  if (ret)
+  {
+    error_no = errno;
+  }
+  goto end;
+#elif defined HAVE_SOLARIS_AFFINITY
+  cpu_set_ptr = (psetid_t*)ndb_cpu_set;
+
+  /* Lock against Solaris processor set */
+  ret= pset_bind(*cpu_set_ptr,
+                 P_LWPID,
+                 pThread->tid,
+                 NULL);
+  if (ret)
+  {
+    error_no = errno;
+    goto end;
+  }
+  return 0;
+#else
+  (void) ret;
+  goto end;
+#endif
+
+end:
+  if (!error_no)
+    pThread->cpu_set_key = cpu_set_key;
+  return error_no;
+}
+
+const struct processor_set_handler*
+NdbThread_LockGetCPUSetKey(struct NdbThread* pThread)
+{
+  return pThread->cpu_set_key;
+}
+
+int
+NdbThread_LockCPU(struct NdbThread* pThread,
+                  Uint32 cpu_id,
+                  const struct processor_set_handler *cpu_set_key)
+{
+  int error_no;
+  int ret;
+#if defined HAVE_LINUX_SCHEDULING
+  cpu_set_t cpu_set;
+#endif
+
+  if ((error_no = set_old_cpu_locking(pThread)))
+    goto end;
+
 #if defined HAVE_LINUX_SCHEDULING
 
   /*
@@ -478,13 +849,14 @@ NdbThread_LockCPU(struct NdbThread* pThread, Uint32 cpu_id)
     By combining Real-time Scheduling and Locking to CPU we can
     achieve more or less a realtime system for NDB Cluster.
   */
-  int ret;
-  cpu_set_t cpu_set;
+
   CPU_ZERO(&cpu_set);
   CPU_SET(cpu_id, &cpu_set);
   ret= sched_setaffinity(pThread->tid, sizeof(cpu_set), &cpu_set);
   if (ret)
+  {
     error_no = errno;
+  }
 #elif defined HAVE_SOLARIS_AFFINITY
   /*
     Solaris have a number of versions to lock threads to CPU's.
@@ -493,13 +865,20 @@ NdbThread_LockCPU(struct NdbThread* pThread, Uint32 cpu_id)
     A bit unclear as whether the id returned by pthread_self
     is the LWP id.
   */
-  int ret;
   ret= processor_bind(P_LWPID, pThread->tid, cpu_id, NULL);
   if (ret)
+  {
     error_no= errno;
+  }
 #else
+  (void) ret;
+  (void) cpu_id;
   error_no = ENOSYS;
+  goto end;
 #endif
+end:
+  if (!error_no)
+    pThread->cpu_set_key = cpu_set_key;
   return error_no;
 }
 
@@ -509,6 +888,11 @@ static pthread_key(void*, tls_keys[NDB_THREAD_TLS_MAX]);
 static pthread_key(void*, tls_keys[NDB_THREAD_TLS_MAX + 1]);
 #endif
 
+struct NdbThread* NdbThread_GetNdbThread()
+{
+  return (struct NdbThread*)NdbThread_GetTlsKey(NDB_THREAD_TLS_NDB_THREAD);
+}
+
 void *NdbThread_GetTlsKey(NDB_THREAD_TLS key)
 {
   return pthread_getspecific(tls_keys[key]);
@@ -526,6 +910,7 @@ NdbThread_Init()
   g_ndb_thread_condition = NdbCondition_Create();
   pthread_key_create(&(tls_keys[NDB_THREAD_TLS_JAM]), NULL);
   pthread_key_create(&(tls_keys[NDB_THREAD_TLS_THREAD]), NULL);
+  pthread_key_create(&(tls_keys[NDB_THREAD_TLS_NDB_THREAD]), NULL);
 #ifdef NDB_MUTEX_DEADLOCK_DETECTOR
   pthread_key_create(&(tls_keys[NDB_THREAD_TLS_MAX]), NULL);
 #endif
@@ -548,7 +933,7 @@ NdbThread_End()
 
   if (g_main_thread)
   {
-    NdbMem_Free((char *)g_main_thread);
+    NdbMem_Free(g_main_thread);
     g_main_thread = 0;
   }
 }
