@@ -1,14 +1,21 @@
 /*
-   Copyright (c) 2003, 2011, Oracle and/or its affiliates. All rights reserved.
+   Copyright (c) 2003, 2015, Oracle and/or its affiliates. All rights reserved.
 
    This program is free software; you can redistribute it and/or modify
-   it under the terms of the GNU General Public License as published by
-   the Free Software Foundation; version 2 of the License.
+   it under the terms of the GNU General Public License, version 2.0,
+   as published by the Free Software Foundation.
+
+   This program is also distributed with certain software (including
+   but not limited to OpenSSL) that is licensed under separate terms,
+   as designated in a particular file or component or in included license
+   documentation.  The authors of MySQL hereby grant you an additional
+   permission to link the program and your derivative works with the
+   separately licensed software that they have included with MySQL.
 
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-   GNU General Public License for more details.
+   GNU General Public License, version 2.0, for more details.
 
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
@@ -39,7 +46,7 @@
 #include <ErrorReporter.hpp>
 #include <ErrorHandlingMacros.hpp>
 
-#include "DLList.hpp"
+#include "IntrusiveList.hpp"
 #include "ArrayPool.hpp"
 #include "DLHashTable.hpp"
 #include "WOPool.hpp"
@@ -54,10 +61,29 @@
 #include <Configuration.hpp>
 
 #include <signaldata/ReadConfig.hpp>
-#include <signaldata/UpgradeStartup.hpp>
 #include "ndbd_malloc_impl.hpp"
 #include <blocks/record_types.hpp>
 
+#include "Ndbinfo.hpp"
+
+#define JAM_FILE_ID 248
+
+
+#ifdef VM_TRACE
+#define D(x) \
+  do { \
+    char buf[200]; \
+    if (!debugOutOn()) break; \
+    debugOutLock(); \
+    debugOutStream() << debugOutTag(buf, __LINE__) << x << dec << "\n"; \
+    debugOutUnlock(); \
+  } while (0)
+#define V(x) " " << #x << ":" << (x)
+#else
+#define D(x) do { } while(0)
+#undef V
+#endif
+
 /**
  * Something for filesystem access
  */
@@ -87,20 +113,32 @@ struct Block_context
   class Ndbd_mem_manager& m_mm;
 };
 
-class SimulatedBlock {
+struct PackedWordsContainer
+{
+  BlockReference hostBlockRef;
+  Uint32 noOfPackedWords;
+  Uint32 packedWords[30];
+}; // 128 bytes
+class SimulatedBlock :
+  public SegmentUtils  /* SimulatedBlock implements the Interface */
+{
   friend class TraceLCP;
   friend class SafeCounter;
   friend class SafeCounterManager;
-  friend struct UpgradeStartup;
   friend class AsyncFile;
   friend class PosixAsyncFile; // FIXME
+  friend class Win32AsyncFile;
   friend class Pgman;
   friend class Page_cache_client;
   friend class Lgman;
   friend class Logfile_client;
+  friend class Tablespace_client;
+  friend class Dbtup_client;
   friend struct Pool_context;
   friend struct SectionHandle;
   friend class LockQueue;
+  friend class SimplePropertiesSectionWriter;
+  friend class SegmentedSectionGuard;
 public:
   friend class BlockComponent;
   virtual ~SimulatedBlock();
@@ -110,7 +148,8 @@ protected:
    * Constructor
    */
   SimulatedBlock(BlockNumber blockNumber,
-		 struct Block_context & ctx); 
+		 struct Block_context & ctx,
+                 Uint32 instanceNumber = 0);
   
   /**********************************************************
    * Handling of execFunctions
@@ -122,19 +161,119 @@ protected:
 
   void initCommon();
 public:
-  /**
-   * 
-   */
-  inline void executeFunction(GlobalSignalNumber gsn, Signal* signal);
-public:
-  typedef void (SimulatedBlock::* CallbackFunction)(class Signal*, 
+  typedef void (SimulatedBlock::* CallbackFunction)(class Signal*,
 						    Uint32 callbackData,
 						    Uint32 returnCode);
   struct Callback {
     CallbackFunction m_callbackFunction;
     Uint32 m_callbackData;
   };
+
+  /**
+   * 
+   */
+  inline void executeFunction(GlobalSignalNumber gsn, Signal* signal);
+
+  /* Multiple block instances */
+  Uint32 instance() const {
+    return theInstance;
+  }
+  SimulatedBlock* getInstance(Uint32 instanceNumber) {
+    ndbrequire(theInstance == 0); // valid only on main instance
+    if (instanceNumber == 0)
+      return this;
+    ndbrequire(instanceNumber < MaxInstances);
+    if (theInstanceList != 0)
+      return theInstanceList[instanceNumber];
+    return 0;
+  }
+  void addInstance(SimulatedBlock* b, Uint32 theInstanceNo);
+  virtual void loadWorkers() {}
+
+  struct ThreadContext
+  {
+    Uint32 threadId;
+    EmulatedJamBuffer* jamBuffer;
+    Uint32 * watchDogCounter;
+    SectionSegmentPool::Cache * sectionPoolCache;
+  };
+  /* Setup state of a block object for executing in a particular thread. */
+  void assignToThread(ThreadContext ctx);
+  /* For multithreaded ndbd, get the id of owning thread. */
+  uint32 getThreadId() const { return m_threadId; }
+  static bool isMultiThreaded();
+
+  /* Configuration based alternative.  Applies only to this node */
+  static bool isNdbMt() { return globalData.isNdbMt; }
+  static bool isNdbMtLqh() { return globalData.isNdbMtLqh; }
+  static Uint32 getLqhWorkers() { return globalData.ndbMtLqhWorkers; }
+
+  /**
+   * Assert that thread calling this function is "owner" of block instance
+   */
+#ifdef VM_TRACE
+  void assertOwnThread();
+#else
+  void assertOwnThread(){ }
+#endif
+
+  /*
+   * Instance key (1-4) is used only when sending a signal.  Receiver
+   * maps it to actual instance (0, if receiver is not MT LQH).
+   *
+   * For performance reason, DBTC gets instance key directly from DBDIH
+   * via DI*GET*NODES*REQ signals.
+   */
+  static Uint32 getInstanceKey(Uint32 tabId, Uint32 fragId);
+  static Uint32 getInstanceFromKey(Uint32 instanceKey); // local use only
+
+  /**
+   * This method will make sure that when callback in called each
+   *   thread running an instance any of the threads in blocks[]
+   *   will have executed a signal
+   */
+  void synchronize_threads_for_blocks(Signal*, const Uint32 blocks[],
+                                      const Callback&, JobBufferLevel = JBB);
   
+  /**
+   * This method make sure that the path specified in blocks[]
+   *   will be traversed before returning
+   */
+  void synchronize_path(Signal*, const Uint32 blocks[],
+                        const Callback&, JobBufferLevel = JBB);
+
+  /**
+   * These methods are used to assist blocks to use the TIME_SIGNAL to
+   * generate drum beats with a regular delay. elapsed_time will report
+   * back the elapsed time since last call but will never report more
+   * than max delay. max_delay = 2 * delay here.
+   */
+  void init_elapsed_time(Signal *signal,
+                         NDB_TICKS &latestTIME_SIGNAL);
+  void sendTIME_SIGNAL(Signal *signal,
+                       const NDB_TICKS currentTime,
+                       Uint32 delay);
+  Uint64 elapsed_time(Signal *signal,
+                      const NDB_TICKS currentTime,
+                      NDB_TICKS &latestTIME_SIGNAL,
+                      Uint32 max_delay);
+
+private:
+  struct SyncThreadRecord
+  {
+    Callback m_callback;
+    Uint32 m_cnt;
+    Uint32 nextPool;
+  };
+  ArrayPool<SyncThreadRecord> c_syncThreadPool;
+  void execSYNC_THREAD_REQ(Signal*);
+  void execSYNC_THREAD_CONF(Signal*);
+
+  void execSYNC_REQ(Signal*);
+
+  void execSYNC_PATH_REQ(Signal*);
+  void execSYNC_PATH_CONF(Signal*);
+public:
   virtual const char* get_filename(Uint32 fd) const { return "";}
 protected:
   static Callback TheEmptyCallback;
@@ -189,6 +328,24 @@ protected:
 		  LinearSectionPtr ptr[3],
 		  Uint32 noOfSections) const ;
 
+  /* NoRelease sendSignal variants do not release sections as
+   * a side-effect of sending.  This requires extra
+   * copying for local sends
+   */
+  void sendSignalNoRelease(BlockReference ref, 
+                           GlobalSignalNumber gsn, 
+                           Signal* signal, 
+                           Uint32 length, 
+                           JobBufferLevel jbuf,
+                           SectionHandle* sections) const;
+
+  void sendSignalNoRelease(NodeReceiverGroup rg,
+                           GlobalSignalNumber gsn,
+                           Signal* signal,
+                           Uint32 length,
+                           JobBufferLevel jbuf,
+                           SectionHandle* sections) const;
+
   // Send multiple signal with delay. In this VM the jobbufffer level has 
   // no effect on on delayed signals
   //
@@ -205,17 +362,36 @@ protected:
 			   Uint32 length,
 			   SectionHandle* sections) const;
 
+  /*
+   * Instance defaults to instance of sender.  Using explicit
+   * instance argument asserts that the call is thread-safe.
+   */
   void EXECUTE_DIRECT(Uint32 block, 
 		      Uint32 gsn, 
 		      Signal* signal, 
-		      Uint32 len);
+		      Uint32 len,
+                      Uint32 givenInstanceNo = ZNIL);
   
   class SectionSegmentPool& getSectionSegmentPool();
+  void release(SegmentedSectionPtr & ptr);
+  void release(SegmentedSectionPtrPOD & ptr) {
+    SegmentedSectionPtr tmp(ptr);
+    release(tmp);
+    ptr.setNull();
+  }
+  void releaseSection(Uint32 firstSegmentIVal);
   void releaseSections(struct SectionHandle&);
 
-  void handle_invalid_sections_in_send_signal(Signal*) const;
-  void handle_lingering_sections_after_execute(Signal*) const;
-  void handle_lingering_sections_after_execute(SectionHandle*) const;
+  bool import(Ptr<SectionSegment> & first, const Uint32 * src, Uint32 len);
+  bool import(SegmentedSectionPtr& ptr, const Uint32* src, Uint32 len);
+  bool import(SectionHandle * dst, LinearSectionPtr src[3],Uint32 cnt);
+
+  bool appendToSection(Uint32& firstSegmentIVal, const Uint32* src, Uint32 len);
+  bool dupSection(Uint32& copyFirstIVal, Uint32 srcFirstIVal);
+  bool writeToSection(Uint32 firstSegmentIVal, Uint32 offset, const Uint32* src, Uint32 len);
+
+  void handle_invalid_sections_in_send_signal(const Signal*) const;
+  void handle_lingering_sections_after_execute(const Signal*) const;
   void handle_invalid_fragmentInfo(Signal*) const;
   void handle_send_failed(SendStatus, Signal*) const;
   void handle_out_of_longsignal_memory(Signal*) const;
@@ -259,6 +435,26 @@ protected:
    */
   bool assembleFragments(Signal * signal);
   
+  /**
+   * Assemble dropped fragments
+   *
+   * Should be called at the start of a Dropped Signal Report 
+   * (GSN_DROPPED_SIGNAL_REP) handler when it is expected that
+   * the block could receive fragmented signals.
+   * No dropped signal handling should be done until this method
+   * returns true.
+   * 
+   * @return true if all fragments has arrived and dropped signal
+   *              handling can proceed.
+   *         false otherwise
+   */
+  bool assembleDroppedFragments(Signal * signal);
+  
+  /* If send size is > FRAGMENT_WORD_SIZE, fragments of this size
+   * will be sent by the sendFragmentedSignal variants
+   */
+  STATIC_CONST( FRAGMENT_WORD_SIZE = 240 );
+
   void sendFragmentedSignal(BlockReference ref, 
 			    GlobalSignalNumber gsn, 
 			    Signal* signal, 
@@ -266,7 +462,7 @@ protected:
 			    JobBufferLevel jbuf,
 			    SectionHandle * sections,
 			    Callback & = TheEmptyCallback,
-			    Uint32 messageSize = 240);
+			    Uint32 messageSize = FRAGMENT_WORD_SIZE);
 
   void sendFragmentedSignal(NodeReceiverGroup rg, 
 			    GlobalSignalNumber gsn, 
@@ -275,7 +471,7 @@ protected:
 			    JobBufferLevel jbuf,
 			    SectionHandle * sections,
 			    Callback & = TheEmptyCallback,
-			    Uint32 messageSize = 240);
+			    Uint32 messageSize = FRAGMENT_WORD_SIZE);
 
   void sendFragmentedSignal(BlockReference ref, 
 			    GlobalSignalNumber gsn, 
@@ -285,7 +481,7 @@ protected:
 			    LinearSectionPtr ptr[3],
 			    Uint32 noOfSections,
 			    Callback & = TheEmptyCallback,
-			    Uint32 messageSize = 240);
+			    Uint32 messageSize = FRAGMENT_WORD_SIZE);
 
   void sendFragmentedSignal(NodeReceiverGroup rg, 
 			    GlobalSignalNumber gsn, 
@@ -295,7 +491,7 @@ protected:
 			    LinearSectionPtr ptr[3],
 			    Uint32 noOfSections,
 			    Callback & = TheEmptyCallback,
-			    Uint32 messageSize = 240);
+			    Uint32 messageSize = FRAGMENT_WORD_SIZE);
 
   /**
    * simBlockNodeFailure
@@ -347,13 +543,20 @@ protected:
     };
     Uint32 prevHash;
     
-    inline bool equal(FragmentInfo & p) const {
+    inline bool equal(FragmentInfo const & p) const {
       return m_senderRef == p.m_senderRef && m_fragmentId == p.m_fragmentId;
     }
     
     inline Uint32 hashValue() const {
       return m_senderRef + m_fragmentId ;
     }
+
+    inline bool isDropped() const {
+      /* IsDropped when entry in hash, but no segments stored */
+      return (( m_sectionPtrI[0] == RNIL ) &&
+              ( m_sectionPtrI[1] == RNIL ) &&
+              ( m_sectionPtrI[2] == RNIL ) );
+    }
   }; // sizeof() = 32 bytes
   
   /**
@@ -369,7 +572,11 @@ protected:
     };
     Uint8  m_status;
     Uint8  m_prio;
-    Uint16 m_fragInfo;
+    Uint8  m_fragInfo;
+    enum Flags {
+      SendNoReleaseSeg = 0x1
+    };
+    Uint8  m_flags;
     Uint16 m_gsn;
     Uint16 m_messageSize; // Size of each fragment
     Uint32 m_fragmentId;
@@ -388,9 +595,11 @@ protected:
   };
   
   /**
-   * setupFragmentSendInfo
-   *   Setup a struct to be used with sendSignalFragment
+   * sendFirstFragment
    *   Used by sendFragmentedSignal
+   *   noRelease can only be used if the caller can guarantee
+   *   not to free the supplied sections until all fragments 
+   *   have been sent.
    */
   bool sendFirstFragment(FragmentSendInfo & info,
 			 NodeReceiverGroup rg, 
@@ -399,7 +608,8 @@ protected:
 			 Uint32 length, 
 			 JobBufferLevel jbuf,
 			 SectionHandle * sections,
-			 Uint32 messageSize = 240);
+                         bool noRelease,
+			 Uint32 messageSize = FRAGMENT_WORD_SIZE);
   
   bool sendFirstFragment(FragmentSendInfo & info,
 			 NodeReceiverGroup rg, 
@@ -409,7 +619,7 @@ protected:
 			 JobBufferLevel jbuf,
 			 LinearSectionPtr ptr[3],
 			 Uint32 noOfSections,
-			 Uint32 messageSize = 240);
+			 Uint32 messageSize = FRAGMENT_WORD_SIZE);
   
   /**
    * Send signal fragment
@@ -426,6 +636,10 @@ protected:
   void sendNextLinearFragment(Signal* signal, FragmentSendInfo & info);
   
   BlockNumber    number() const;
+public:
+  /* Must be public so that we can jam() outside of block scope. */
+  EmulatedJamBuffer *jamBuffer() const;
+protected:
   BlockReference reference() const;
   NodeId         getOwnNodeId() const;
 
@@ -449,8 +663,35 @@ private:
     ATTRIBUTE_NORETURN;
   const NodeId         theNodeId;
   const BlockNumber    theNumber;
+  const Uint32 theInstance;
   const BlockReference theReference;
+  /*
+   * Instance 0 is the main instance.  It creates/owns other instances.
+   * In MT LQH main instance is the LQH proxy and the others ("workers")
+   * are real LQHs run by multiple threads.
+   */
+protected:
+  enum { MaxInstances = NDBMT_MAX_BLOCK_INSTANCES };
+private:
+  SimulatedBlock** theInstanceList; // set in main, indexed by instance
+  SimulatedBlock* theMainInstance;  // set in all
+  /*
+    Thread id currently executing this block.
+    Not used in singlethreaded ndbd.
+  */
+  Uint32 m_threadId;
+  /*
+    Jam buffer reference.
+    In multithreaded ndbd, this is different in each thread, and must be
+    updated if migrating the block to another thread.
+  */
+  EmulatedJamBuffer *m_jamBuffer;
+  /* For multithreaded ndb, the thread-specific watchdog counter. */
+  Uint32 *m_watchDogCounter;
 
+  SectionSegmentPool::Cache * m_sectionPoolCache;
+  
+  
   Uint32 doNodeFailureCleanup(Signal* signal,
                               Uint32 failedNodeId,
                               Uint32 resource,
@@ -472,8 +713,10 @@ protected:
   Block_context m_ctx;
   NewVARIABLE* allocateBat(int batSize);
   void freeBat();
-  static const NewVARIABLE* getBat    (BlockNumber blockNo);
-  static Uint16             getBatSize(BlockNumber blockNo);
+  static const NewVARIABLE* getBat    (BlockNumber blockNo,
+                                       Uint32 instanceNo);
+  static Uint16             getBatSize(BlockNumber blockNo,
+                                       Uint32 instanceNo);
   
   static BlockReference calcTcBlockRef   (NodeId aNode);
   static BlockReference calcLqhBlockRef  (NodeId aNode);
@@ -490,6 +733,13 @@ protected:
 
   static BlockReference calcApiClusterMgrBlockRef (NodeId aNode);
 
+  // matching instance on same node e.g. LQH-ACC-TUP
+  BlockReference calcInstanceBlockRef(BlockNumber aBlock);
+
+  // matching instance on another node e.g. LQH-LQH
+  // valid only if receiver has same number of workers
+  BlockReference calcInstanceBlockRef(BlockNumber aBlock, NodeId aNode);
+
   /** 
    * allocRecord
    * Allocates memory for the datastructures where ndb keeps the data
@@ -565,7 +815,8 @@ protected:
    *
    */
   Uint32 create_distr_key(Uint32 tableId,
-			  Uint32 *data, 
+			  const Uint32* src,
+                          Uint32 *dst, 
 			  const Uint32 keyPaLen[MAX_ATTRIBUTES_IN_INDEX])const;
   
   /**
@@ -581,12 +832,18 @@ protected:
    */
   void setup_wakeup();
 
+  /**
+   * Get receiver thread index for node
+   * MAX_NODES == no receiver thread
+   */
+  Uint32 get_recv_thread_idx(NodeId nodeId);
+
 private:
   NewVARIABLE* NewVarRef;      /* New Base Address Table for block  */
   Uint16       theBATSize;     /* # entries in BAT */
 
 protected:  
-  ArrayPool<GlobalPage>& m_global_page_pool;
+  SafeArrayPool<GlobalPage>& m_global_page_pool;
   ArrayPool<GlobalPage>& m_shared_page_pool;
   
   void execNDB_TAMPER(Signal * signal);
@@ -595,9 +852,10 @@ protected:
 
   void execSIGNAL_DROPPED_REP(Signal* signal);
   void execCONTINUE_FRAGMENTED(Signal* signal);
+  void execSTOP_FOR_CRASH(Signal* signal);
   void execAPI_START_REP(Signal* signal);
   void execNODE_START_REP(Signal* signal);
-
+  void execSEND_PACKED(Signal* signal);
   void execLOCAL_ROUTE_ORD(Signal*);
 private:
   /**
@@ -690,8 +948,10 @@ private:
   void execUTIL_UNLOCK_REF(Signal* signal);
   void execUTIL_UNLOCK_CONF(Signal* signal);
 
+  void check_sections(Signal* signal, 
+                      Uint32 oldSecCount, 
+                      Uint32 newSecCount) const;
 protected:
-  void execUPGRADE(Signal* signal);
 
   void fsRefError(Signal* signal, Uint32 line, const char *msg);
   void execFSWRITEREF(Signal* signal);
@@ -702,6 +962,42 @@ protected:
   void execFSSYNCREF(Signal* signal);
   void execFSAPPENDREF(Signal* signal);
 
+  // MT LQH callback CONF via signal
+public:
+  struct CallbackPtr {
+    Uint32 m_callbackIndex;
+    Uint32 m_callbackData;
+  };
+protected:
+  enum CallbackFlags {
+    CALLBACK_DIRECT = 0x0001, // use EXECUTE_DIRECT (assumed thread safe)
+    CALLBACK_ACK    = 0x0002  // send ack at the end of callback timeslice
+  };
+
+  struct CallbackEntry {
+    CallbackFunction m_function;
+    Uint32 m_flags;
+  };
+
+  struct CallbackTable {
+    Uint32 m_count;
+    CallbackEntry* m_entry; // array
+  };
+
+  CallbackTable* m_callbackTableAddr; // set by block if used
+
+  enum {
+    THE_NULL_CALLBACK = 0 // must assign TheNULLCallbackFunction
+  };
+
+  void execute(Signal* signal, CallbackPtr & cptr, Uint32 returnCode);
+  const CallbackEntry& getCallbackEntry(Uint32 ci);
+  void sendCallbackConf(Signal* signal, Uint32 fullBlockNo,
+                        CallbackPtr& cptr,
+                        Uint32 senderData, Uint32 callbackInfo,
+                        Uint32 returnCode);
+  void execCALLBACK_CONF(Signal* signal);
+
   // Variable for storing inserted errors, see pc.H
   ERROR_INSERT_VARIABLE;
 
@@ -725,8 +1021,67 @@ public:
   void disable_global_variables();
   void enable_global_variables();
 #endif
+
+#ifdef VM_TRACE
+public:
+  NdbOut debugOut;
+  NdbOut& debugOutStream() { return debugOut; };
+  bool debugOutOn();
+  void debugOutLock() { globalSignalLoggers.lock(); }
+  void debugOutUnlock() { globalSignalLoggers.unlock(); }
+  const char* debugOutTag(char* buf, int line);
+#endif
+
+  void ndbinfo_send_row(Signal* signal,
+                        const DbinfoScanReq& req,
+                        const Ndbinfo::Row& row,
+                        Ndbinfo::Ratelimit& rl) const;
+
+  void ndbinfo_send_scan_break(Signal* signal,
+                               DbinfoScanReq& req,
+                               const Ndbinfo::Ratelimit& rl,
+                               Uint32 data1, Uint32 data2 = 0,
+                               Uint32 data3 = 0, Uint32 data4 = 0) const;
+
+  void ndbinfo_send_scan_conf(Signal* signal,
+                              DbinfoScanReq& req,
+                              const Ndbinfo::Ratelimit& rl) const;
+
+
+protected:
+  /**
+   * SegmentUtils methods
+   */
+  virtual SectionSegment* getSegmentPtr(Uint32 iVal);
+  virtual bool seizeSegment(Ptr<SectionSegment>& p);
+  virtual void releaseSegment(Uint32 iVal);
+
+  virtual void releaseSegmentList(Uint32 firstSegmentIVal);
+
+  /** End of SegmentUtils methods */
 };
 
+// outside blocks e.g. within a struct
+#ifdef VM_TRACE
+#define DEBUG_OUT_DEFINES(blockNo) \
+static SimulatedBlock* debugOutBlock() \
+  { return globalData.getBlock(blockNo); } \
+static NdbOut& debugOutStream() \
+  { return debugOutBlock()->debugOutStream(); } \
+static bool debugOutOn() \
+  { return debugOutBlock()->debugOutOn(); } \
+static void debugOutLock() \
+  { debugOutBlock()->debugOutLock(); } \
+static void debugOutUnlock() \
+  { debugOutBlock()->debugOutUnlock(); } \
+static const char* debugOutTag(char* buf, int line) \
+  { return debugOutBlock()->debugOutTag(buf, line); } \
+static void debugOutDefines()
+#else
+#define DEBUG_OUT_DEFINES(blockNo) \
+static void debugOutDefines()
+#endif
+
 inline 
 void 
 SimulatedBlock::executeFunction(GlobalSignalNumber gsn, Signal* signal){
@@ -770,12 +1125,29 @@ SimulatedBlock::execute(Signal* signal, Callback & c, Uint32 returnCode){
   (this->*fun)(signal, c.m_callbackData, returnCode);
 }
 
+inline
+void
+SimulatedBlock::execute(Signal* signal, CallbackPtr & cptr, Uint32 returnCode){
+  const CallbackEntry& ce = getCallbackEntry(cptr.m_callbackIndex);
+  cptr.m_callbackIndex = ZNIL;
+  Callback c;
+  c.m_callbackFunction = ce.m_function;
+  c.m_callbackData = cptr.m_callbackData;
+  execute(signal, c, returnCode);
+}
+                        
 inline 
 BlockNumber
 SimulatedBlock::number() const {
    return theNumber;
 }
 
+inline
+EmulatedJamBuffer *
+SimulatedBlock::jamBuffer() const {
+   return m_jamBuffer;
+}
+
 inline
 BlockReference
 SimulatedBlock::reference() const {
@@ -866,6 +1238,18 @@ SimulatedBlock::calcApiClusterMgrBlockRef (NodeId aNodeId){
   return numberToRef(API_CLUSTERMGR, aNodeId);
 }
 
+inline
+BlockReference
+SimulatedBlock::calcInstanceBlockRef(BlockNumber aBlock){
+  return numberToRef(aBlock, instance(), getOwnNodeId());
+}
+
+inline
+BlockReference
+SimulatedBlock::calcInstanceBlockRef(BlockNumber aBlock, NodeId aNodeId){
+  return numberToRef(aBlock, instance(), aNodeId);
+}
+
 inline
 const NodeState &
 SimulatedBlock::getNodeState() const {
@@ -891,57 +1275,6 @@ SimulatedBlock::setNodeVersionInfo() {
   return globalData.m_versionInfo;
 }
 
-inline
-void
-SimulatedBlock::EXECUTE_DIRECT(Uint32 block, 
-			       Uint32 gsn, 
-			       Signal* signal, 
-			       Uint32 len){
-  signal->setLength(len);
-#ifdef VM_TRACE
-  if(globalData.testOn){
-    signal->header.theVerId_signalNumber = gsn;
-    signal->header.theReceiversBlockNumber = block;
-    signal->header.theSendersBlockRef = reference();
-    globalSignalLoggers.executeDirect(signal->header,
-				      0,        // in
-				      &signal->theData[0],
-                                      globalData.ownId);
-  }
-#endif
-  SimulatedBlock* b = globalData.getBlock(block);
-#ifdef VM_TRACE_TIME
-  Uint32 us1, us2;
-  Uint64 ms1, ms2;
-  NdbTick_CurrentMicrosecond(&ms1, &us1);
-  Uint32 tGsn = m_currentGsn;
-  b->m_currentGsn = gsn;
-#endif
-  b->executeFunction(gsn, signal);
-#ifdef VM_TRACE_TIME
-  NdbTick_CurrentMicrosecond(&ms2, &us2);
-  Uint64 diff = ms2;
-  diff -= ms1;
-  diff *= 1000000;
-  diff += us2;
-  diff -= us1;
-  b->addTime(gsn, diff);
-  m_currentGsn = tGsn;
-  subTime(tGsn, diff);
-#endif
-#ifdef VM_TRACE
-  if(globalData.testOn){
-    signal->header.theVerId_signalNumber = gsn;
-    signal->header.theReceiversBlockNumber = block;
-    signal->header.theSendersBlockRef = reference();
-    globalSignalLoggers.executeDirect(signal->header,
-				      1,        // out
-				      &signal->theData[0],
-                                      globalData.ownId);
-  }
-#endif
-}
-
 #ifdef VM_TRACE_TIME
 inline
 void
@@ -957,6 +1290,78 @@ SimulatedBlock::subTime(Uint32 gsn, Uint64 time){
 }
 #endif
 
+inline
+void
+SimulatedBlock::EXECUTE_DIRECT(Uint32 block, 
+			       Uint32 gsn, 
+			       Signal* signal, 
+			       Uint32 len,
+                               Uint32 givenInstanceNo)
+{
+  signal->setLength(len);
+  SimulatedBlock* b = globalData.getBlock(block);
+  ndbassert(b != 0);
+  /**
+   * In multithreaded NDB, blocks run in different threads, and EXECUTE_DIRECT
+   * (unlike sendSignal) is generally not thread-safe.
+   * So only allow EXECUTE_DIRECT between blocks that run in the same thread,
+   * unless caller explicitly marks it as being thread safe (eg NDBFS),
+   * by using an explicit instance argument.
+   * By default instance of sender is used.  This is automatically thread-safe
+   * for worker instances (instance != 0).
+   */
+  Uint32 instanceNo = givenInstanceNo;
+  if (instanceNo == ZNIL)
+    instanceNo = instance();
+  if (instanceNo != 0)
+    b = b->getInstance(instanceNo);
+  ndbassert(b != 0);
+  ndbassert(givenInstanceNo != ZNIL || b->getThreadId() == getThreadId());
+  signal->header.theSendersBlockRef = reference();
+#ifdef VM_TRACE
+  if(globalData.testOn){
+    signal->header.theVerId_signalNumber = gsn;
+    signal->header.theReceiversBlockNumber = numberToBlock(block, instanceNo);
+    globalSignalLoggers.executeDirect(signal->header,
+				      0,        // in
+				      &signal->theData[0],
+                                      globalData.ownId);
+  }
+#endif
+#ifdef VM_TRACE_TIME
+  const NDB_TICKS t1 = NdbTick_getCurrentTicks();
+  Uint32 tGsn = m_currentGsn;
+  b->m_currentGsn = gsn;
+#endif
+  b->executeFunction(gsn, signal);
+#ifdef VM_TRACE_TIME
+  const NDB_TICKS t2 = NdbTick_getCurrentTicks();
+  const Uint64 diff = NdbTick_Elapsed(t1,t2).microSec();
+  b->addTime(gsn, diff);
+  m_currentGsn = tGsn;
+  subTime(tGsn, diff);
+#endif
+}
+
+// Do a consictency check before reusing a signal.
+inline void 
+SimulatedBlock::check_sections(Signal* signal, 
+                               Uint32 oldSecCount, 
+                               Uint32 newSecCount) const
+{ 
+  // Sections from previous use should have been consumed by now.
+  if (unlikely(oldSecCount != 0))
+  { 
+    handle_invalid_sections_in_send_signal(signal); 
+  } 
+  else if (unlikely(newSecCount == 0 &&
+                    signal->header.m_fragmentInfo != 0 && 
+                    signal->header.m_fragmentInfo != 3))
+  { 
+    handle_invalid_fragmentInfo(signal); 
+  }
+}
+
 /**
  * Defines for backward compatiblility
  */
@@ -979,17 +1384,6 @@ BLOCK::addRecSignal(GlobalSignalNumber gsn, ExecSignalLocal f, bool force){ \
   addRecSignalImpl(gsn, (ExecFunction)f, force);\
 }
 
-#include "Mutex.hpp"
-
-inline
-SectionHandle::~SectionHandle()
-{
-  if (unlikely(m_cnt))
-  {
-    m_block->handle_lingering_sections_after_execute(this);
-  }
-}
-
 #ifdef ERROR_INSERT
 #define RSS_AP_SNAPSHOT(x) Uint32 rss_##x
 #define RSS_AP_SNAPSHOT_SAVE(x) rss_##x = x.getNoOfFree()
@@ -1007,6 +1401,10 @@ SectionHandle::~SectionHandle()
 #define RSS_OP_SNAPSHOT(x) Uint32 rss_##x
 #define RSS_OP_SNAPSHOT_SAVE(x) rss_##x = x
 #define RSS_OP_SNAPSHOT_CHECK(x) ndbrequire(rss_##x == x)
+
+#define RSS_DA256_SNAPSHOT(x) Uint32 rss_##x
+#define RSS_DA256_SNAPSHOT_SAVE(x) rss_##x = x.m_high_pos
+#define RSS_DA256_SNAPSHOT_CHECK(x) ndbrequire(x.m_high_pos <= rss_##x)
 #else
 #define RSS_AP_SNAPSHOT(x) struct rss_dummy0_##x { int dummy; }
 #define RSS_AP_SNAPSHOT_SAVE(x)
@@ -1025,7 +1423,65 @@ SectionHandle::~SectionHandle()
 #define RSS_OP_SNAPSHOT_SAVE(x)
 #define RSS_OP_SNAPSHOT_CHECK(x)
 
+#define RSS_DA256_SNAPSHOT(x)
+#define RSS_DA256_SNAPSHOT_SAVE(x)
+#define RSS_DA256_SNAPSHOT_CHECK(x)
 #endif
 
+struct Hash2FragmentMap
+{
+  STATIC_CONST( MAX_MAP = NDB_MAX_HASHMAP_BUCKETS );
+  Uint32 m_cnt;
+  Uint32 m_fragments;
+  Uint16 m_map[MAX_MAP];
+  Uint32 nextPool;
+  Uint32 m_object_id;
+};
+
+extern ArrayPool<Hash2FragmentMap> g_hash_map;
+
+/**
+ * Guard class for auto release of segmentedsectionptr's
+ */
+class SegmentedSectionGuard
+{
+  Uint32 cnt;
+  Uint32 ptr[3];
+  SimulatedBlock * block;
+
+public:
+  SegmentedSectionGuard(SimulatedBlock* b) : cnt(0), block(b) { }
+  SegmentedSectionGuard(SimulatedBlock* b, Uint32 ptrI) : cnt(1), block(b) {
+    ptr[0] = ptrI;
+  }
+
+  void add(Uint32 ptrI) {
+    if (ptrI != RNIL)
+    {
+      assert(cnt < NDB_ARRAY_SIZE(ptr));
+      ptr[cnt] = ptrI;
+      cnt++;
+    }
+  }
+
+  void release() {
+    for (Uint32 i = 0; i < cnt; i++) {
+      if (ptr[i] != RNIL)
+        block->releaseSection(ptr[i]);
+    }
+    cnt = 0;
+  }
+
+  void clear() {
+    cnt = 0;
+  }
+
+  ~SegmentedSectionGuard() {
+    release();
+  }
+};
+
+#undef JAM_FILE_ID
+
 #endif
 
