@@ -1,27 +1,30 @@
-/* Copyright (C) 2003 MySQL AB
+/*
+   Copyright (c) 2003, 2013, Oracle and/or its affiliates. All rights reserved.
 
    This program is free software; you can redistribute it and/or modify
-   it under the terms of the GNU General Public License as published by
-   the Free Software Foundation; version 2 of the License.
+   it under the terms of the GNU General Public License, version 2.0,
+   as published by the Free Software Foundation.
+
+   This program is also distributed with certain software (including
+   but not limited to OpenSSL) that is licensed under separate terms,
+   as designated in a particular file or component or in included license
+   documentation.  The authors of MySQL hereby grant you an additional
+   permission to link the program and your derivative works with the
+   separately licensed software that they have included with MySQL.
 
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-   GNU General Public License for more details.
+   GNU General Public License, version 2.0, for more details.
 
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
-   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */
+   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA
+*/
 
-#include <ndb_global.h>
-#include <NdbOperation.hpp>
-#include "NdbApiSignal.hpp"
-#include <NdbTransaction.hpp>
-#include <Ndb.hpp>
-#include <NdbRecAttr.hpp>
-#include "NdbUtil.hpp"
+#include "API.hpp"
 #include "Interpreter.hpp"
-#include <NdbIndexScanOperation.hpp>
+#include <signaldata/AttrInfo.hpp>
 
 #ifdef VM_TRACE
 #include <NdbEnv.h>
@@ -54,25 +57,40 @@ NdbOperation::initInterpreter(){
   theFinalReadSize = 0;
   theInterpretIndicator = 1;
 
-  theTotalCurrAI_Len = 5;
+  theTotalCurrAI_Len = AttrInfo::SectionSizeInfoLength;
+}
+
+bool
+NdbOperation::isNdbRecordOperation()
+{
+  /* All scans are 'NdbRecord'.  For PK and UK access
+   * check if we've got an m_attribute_record set 
+   */
+  return !(((m_type == PrimaryKeyAccess) ||
+            (m_type == UniqueIndexAccess)) &&
+           (m_attribute_record == NULL));
 }
 
 int
 NdbOperation::incCheck(const NdbColumnImpl* tNdbColumnImpl)
 {
+  if (isNdbRecordOperation()) {
+    /* Wrong API.  Use NdbInterpretedCode for NdbRecord operations */
+    setErrorCodeAbort(4537);
+    return -1;
+  }
+
   if ((theInterpretIndicator == 1)) {
-    if ((tNdbColumnImpl == NULL) ||
-        (theOperationType == OpenScanRequest ||
-         theOperationType == OpenRangeScanRequest))
+    if (tNdbColumnImpl == NULL)
       goto inc_check_error1;
     if ((tNdbColumnImpl->getInterpretableType() != true) ||
         (tNdbColumnImpl->m_pk != false) ||
         (tNdbColumnImpl->m_nullable))
       goto inc_check_error2;
-    if (theStatus == ExecInterpretedValue || theStatus == UseNdbRecord) {
+    if (theStatus == ExecInterpretedValue) {
       ; // Simply continue with interpretation
     } else if (theStatus == GetValue) {
-      theInitialReadSize = theTotalCurrAI_Len - 5;
+      theInitialReadSize = theTotalCurrAI_Len - AttrInfo::SectionSizeInfoLength;
       theStatus = ExecInterpretedValue;
     } else if (theStatus == SubroutineExec) {
       ; // Simply continue with interpretation
@@ -80,8 +98,10 @@ NdbOperation::incCheck(const NdbColumnImpl* tNdbColumnImpl)
       setErrorCodeAbort(4231);
       return -1;
     }
-    m_no_disk_flag &= 
-      (tNdbColumnImpl->m_storageType == NDB_STORAGETYPE_DISK ? 0:1);
+    if (tNdbColumnImpl->m_storageType == NDB_STORAGETYPE_DISK)
+    {
+      m_flags &= ~(Uint8)OF_NO_DISK;
+    }
     return tNdbColumnImpl->m_attrId;
   } else {
     if (theNdbCon->theCommitStatus == NdbTransaction::Started)
@@ -90,11 +110,6 @@ NdbOperation::incCheck(const NdbColumnImpl* tNdbColumnImpl)
   return -1;
   
  inc_check_error1:
-  if (theOperationType == OpenScanRequest ||
-      theOperationType == OpenRangeScanRequest) {
-    setErrorCodeAbort(4228);
-    return -1;
-  }
   setErrorCodeAbort(4004);
   return -1;
   
@@ -118,15 +133,19 @@ NdbOperation::incCheck(const NdbColumnImpl* tNdbColumnImpl)
 int
 NdbOperation::write_attrCheck(const NdbColumnImpl* tNdbColumnImpl)
 {
+  if (isNdbRecordOperation()) {
+    /* Wrong API.  Use NdbInterpretedCode for NdbRecord operations */
+    setErrorCodeAbort(4537);
+    return -1;
+  }
+
   if ((theInterpretIndicator == 1)) {
-    if ((tNdbColumnImpl == NULL) ||
-        (theOperationType == OpenScanRequest ||
-         theOperationType == OpenRangeScanRequest))
+    if (tNdbColumnImpl == NULL)
       goto write_attr_check_error1;
     if ((tNdbColumnImpl->getInterpretableType() == false) ||
         (tNdbColumnImpl->m_pk))
       goto write_attr_check_error2;
-    if (theStatus == ExecInterpretedValue || theStatus == UseNdbRecord) {
+    if (theStatus == ExecInterpretedValue) {
       ; // Simply continue with interpretation
     } else if (theStatus == SubroutineExec) {
       ; // Simply continue with interpretation
@@ -134,8 +153,10 @@ NdbOperation::write_attrCheck(const NdbColumnImpl* tNdbColumnImpl)
       setErrorCodeAbort(4231);
       return -1;
     }
-    m_no_disk_flag &= 
-      (tNdbColumnImpl->m_storageType == NDB_STORAGETYPE_DISK ? 0:1);
+    if (tNdbColumnImpl->m_storageType == NDB_STORAGETYPE_DISK)
+    {
+      m_flags &= ~(Uint8)OF_NO_DISK;
+    }
     return tNdbColumnImpl->m_attrId;
   } else {
     if (theNdbCon->theCommitStatus == NdbTransaction::Started)
@@ -144,11 +165,6 @@ NdbOperation::write_attrCheck(const NdbColumnImpl* tNdbColumnImpl)
   return -1;
 
 write_attr_check_error1:
-  if (theOperationType == OpenScanRequest ||
-      theOperationType == OpenRangeScanRequest) {
-    setErrorCodeAbort(4228);
-    return -1;
-  }
   setErrorCodeAbort(4004);
   return -1;
 
@@ -168,15 +184,21 @@ write_attr_check_error2:
 int
 NdbOperation::read_attrCheck(const NdbColumnImpl* tNdbColumnImpl)
 {
+  if (isNdbRecordOperation()) {
+    /* Wrong API.  Use NdbInterpretedCode for NdbRecord operations */
+    setErrorCodeAbort(4537);
+    return -1;
+  }
+
   if ((theInterpretIndicator == 1)) {
     if (tNdbColumnImpl == NULL)
       goto read_attr_check_error1;
     if (tNdbColumnImpl->getInterpretableType() == false)
       goto read_attr_check_error2;
-    if (theStatus == ExecInterpretedValue || theStatus == UseNdbRecord) {
+    if (theStatus == ExecInterpretedValue) {
       ; // Simply continue with interpretation
     } else if (theStatus == GetValue) {
-      theInitialReadSize = theTotalCurrAI_Len - 5;
+      theInitialReadSize = theTotalCurrAI_Len - AttrInfo::SectionSizeInfoLength;
       theStatus = ExecInterpretedValue;
     } else if (theStatus == SubroutineExec) {
       ; // Simply continue with interpretation
@@ -184,8 +206,10 @@ NdbOperation::read_attrCheck(const NdbColumnImpl* tNdbColumnImpl)
       setErrorCodeAbort(4231);
       return -1;
     }
-    m_no_disk_flag &= 
-      (tNdbColumnImpl->m_storageType == NDB_STORAGETYPE_DISK ? 0:1);
+    if (tNdbColumnImpl->m_storageType == NDB_STORAGETYPE_DISK)
+    {
+      m_flags &= ~(Uint8)OF_NO_DISK;
+    }
     return tNdbColumnImpl->m_attrId;
   } else {
     if (theNdbCon->theCommitStatus == NdbTransaction::Started)
@@ -209,11 +233,17 @@ NdbOperation::read_attrCheck(const NdbColumnImpl* tNdbColumnImpl)
 int
 NdbOperation::initial_interpreterCheck()
 {
+  if (isNdbRecordOperation()) {
+    /* Wrong API.  Use NdbInterpretedCode for NdbRecord operations */
+    setErrorCodeAbort(4537);
+    return -1;
+  }
+
   if ((theInterpretIndicator == 1)) {
-    if (theStatus == ExecInterpretedValue || theStatus == UseNdbRecord) {
-       return 0; // Simply continue with interpretation
+    if (theStatus == ExecInterpretedValue) {
+      return 0; // Simply continue with interpretation
     } else if (theStatus == GetValue) {
-      theInitialReadSize = theTotalCurrAI_Len - 5;
+      theInitialReadSize = theTotalCurrAI_Len - AttrInfo::SectionSizeInfoLength;
       theStatus = ExecInterpretedValue;
       return 0;
     } else if (theStatus == SubroutineExec) {
@@ -233,11 +263,17 @@ NdbOperation::initial_interpreterCheck()
 int
 NdbOperation::labelCheck()
 {
+  if (isNdbRecordOperation()) {
+    /* Wrong API.  Use NdbInterpretedCode for NdbRecord operations */
+    setErrorCodeAbort(4537);
+    return -1;
+  }
+
   if ((theInterpretIndicator == 1)) {
-    if (theStatus == ExecInterpretedValue || theStatus == UseNdbRecord) {
-       return 0; // Simply continue with interpretation
+    if (theStatus == ExecInterpretedValue) {
+      return 0; // Simply continue with interpretation
     } else if (theStatus == GetValue) {
-      theInitialReadSize = theTotalCurrAI_Len - 5;
+      theInitialReadSize = theTotalCurrAI_Len - AttrInfo::SectionSizeInfoLength;
       theStatus = ExecInterpretedValue;
       return 0;
     } else if (theStatus == SubroutineExec) {
@@ -259,9 +295,15 @@ NdbOperation::labelCheck()
 int
 NdbOperation::intermediate_interpreterCheck()
 {
+  if (isNdbRecordOperation()) {
+    /* Wrong API.  Use NdbInterpretedCode for NdbRecord operations */
+    setErrorCodeAbort(4537);
+    return -1;
+  }
+
   if ((theInterpretIndicator == 1)) {
-    if (theStatus == ExecInterpretedValue || theStatus == UseNdbRecord) {
-       return 0; // Simply continue with interpretation
+    if (theStatus == ExecInterpretedValue) {
+      return 0; // Simply continue with interpretation
     } else if (theStatus == SubroutineExec) {
        return 0; // Simply continue with interpretation
     } else {
@@ -493,11 +535,23 @@ NdbOperation::def_label(int tLabelNo)
    * the NdbOperation::prepareSendInterpreted method.
    */
 
+  Uint32 initialOffset= theInitialReadSize + AttrInfo::SectionSizeInfoLength;
+
+  if (theNoOfSubroutines > 0)
+  {
+    /* Label in a sub, needs to be offset from the start of the subroutines
+     * section
+     */
+    initialOffset+= (theInterpretedSize + theFinalUpdateSize + theFinalReadSize); 
+
+  }
+
   theLastLabel->theLabelNo[tLabelIndex] = tLabelNo;
-  theLastLabel->theLabelAddress[tLabelIndex] = (theTotalCurrAI_Len + 1) - (theInitialReadSize + 5);
+  theLastLabel->theLabelAddress[tLabelIndex] = (theTotalCurrAI_Len + 1) - initialOffset;
   theLastLabel->theSubroutine[tLabelIndex] = theNoOfSubroutines;
   theNoOfLabels++;
   theErrorLine++;
+
   return (theNoOfLabels - 1);
 }
 
@@ -526,7 +580,7 @@ NdbOperation::def_subroutine(int tSubNo)
   {
     theFinalReadSize = theTotalCurrAI_Len -
        (theInitialReadSize + theInterpretedSize + 
-        theFinalUpdateSize + 5);
+        theFinalUpdateSize + AttrInfo::SectionSizeInfoLength);
 
   } else if (theStatus == SubroutineEnd)
   {
@@ -535,16 +589,19 @@ NdbOperation::def_subroutine(int tSubNo)
   {
     if (insertATTRINFO(Interpreter::EXIT_OK) == -1)
       return -1;
-    theInterpretedSize = theTotalCurrAI_Len - (theInitialReadSize + 5);
+    theInterpretedSize = theTotalCurrAI_Len - 
+      (theInitialReadSize + AttrInfo::SectionSizeInfoLength);
   } else if (theStatus == SetValueInterpreted)
   {
     theFinalUpdateSize = theTotalCurrAI_Len -
-       (theInitialReadSize + theInterpretedSize + 5);
+      (theInitialReadSize + theInterpretedSize + 
+       AttrInfo::SectionSizeInfoLength);
 
   } else if (theStatus == GetValue)
   {
 
-    theInitialReadSize = theTotalCurrAI_Len - 5;
+    theInitialReadSize = theTotalCurrAI_Len - 
+      AttrInfo::SectionSizeInfoLength;
 
   } else
   {
@@ -570,8 +627,9 @@ NdbOperation::def_subroutine(int tSubNo)
     tNdbSubroutine->theNext = NULL;
   }
   theLastSubroutine->theSubroutineAddress[tSubroutineIndex] = theTotalCurrAI_Len - 
-    (theInitialReadSize + theInterpretedSize + 
-     theFinalUpdateSize + theFinalReadSize);
+    (AttrInfo::SectionSizeInfoLength + theInitialReadSize + theInterpretedSize + 
+     theFinalUpdateSize + theFinalReadSize); // Preceding sections + sizes array
+
   theNoOfSubroutines++;
   theErrorLine++;
   return (theNoOfSubroutines - 1);
@@ -913,6 +971,10 @@ int
 NdbOperation::interpret_exit_nok()
 {
   INT_DEBUG(("interpret_exit_nok"));
+  /**
+   * 899 is used here for historical reasons. Observe that this collides with 
+   * "Rowid already allocated" (see ndberror.c).
+   */
   Uint32 ErrorCode = 899;
 
   if (initial_interpreterCheck() == -1)
@@ -975,11 +1037,11 @@ NdbOperation::insertBranch(Uint32 aLabel)
   theLastBranch = tBranch;
   if (theNoOfSubroutines == 0)
     tAddress = theTotalCurrAI_Len - 
-       (theInitialReadSize + 5);
+      (theInitialReadSize + AttrInfo::SectionSizeInfoLength);
   else
     tAddress = theTotalCurrAI_Len - 
       (theInitialReadSize + theInterpretedSize +
-       theFinalUpdateSize + theFinalReadSize + 5);
+       theFinalUpdateSize + theFinalReadSize + AttrInfo::SectionSizeInfoLength);
 
   tBranch->theBranchAddress = tAddress;
   tBranch->theSignal = theCurrentATTRINFO;
@@ -1017,7 +1079,7 @@ NdbOperation::insertCall(Uint32 aCall)
 int
 NdbOperation::branch_col(Uint32 type, 
 			 Uint32 ColId, const void * val, Uint32 len, 
-			 bool nopad, Uint32 Label){
+			 Uint32 Label){
 
   DBUG_ENTER("NdbOperation::branch_col");
   DBUG_PRINT("enter", ("type: %u  col:%u  val: 0x%lx  len: %u  label: %u",
@@ -1037,30 +1099,55 @@ NdbOperation::branch_col(Uint32 type,
     abort();
   }
 
+  Uint32 lastWordMask= ~0;
   if (val == NULL)
     len = 0;
   else {
-    if (! col->getStringType()) {
-      // prevent assert in NdbSqlUtil on length error
-      Uint32 sizeInBytes = col->m_attrSize * col->m_arraySize;
-      if (len != 0 && len != sizeInBytes)
+    if (! col->getStringType())
+    {
+      /* Fixed size type */
+      if (col->getType() == NDB_TYPE_BIT)
       {
-        setErrorCodeAbort(4209);
-        DBUG_RETURN(-1);
+        /* We want to zero out insignificant bits in the
+         * last word of a bit type
+         */
+        Uint32 bitLen= col->getLength();
+        Uint32 lastWordBits= bitLen & 0x1F;
+        if (lastWordBits)
+          lastWordMask= (1 << lastWordBits) -1;
+      }
+      len= col->m_attrSize * col->m_arraySize;
+    }
+    else
+    {
+      /* For Like and Not like we must use the passed in 
+       * length.  Otherwise we use the length encoded
+       * in the passed string
+       */
+      if ((type != Interpreter::LIKE) &&
+          (type != Interpreter::NOT_LIKE))
+      {
+        if (! col->get_var_length(val, len))
+        {
+          setErrorCodeAbort(4209);
+          DBUG_RETURN(-1);
+        }
       }
-      len = sizeInBytes;
     }
   }
 
-  m_no_disk_flag &= (col->m_storageType == NDB_STORAGETYPE_DISK ? 0:1);
+  if (col->m_storageType == NDB_STORAGETYPE_DISK)
+  {
+    m_flags &= ~(Uint8)OF_NO_DISK;
+  }
 
-  Uint32 tempData[2000];
+  Uint32 tempData[ NDB_MAX_TUPLE_SIZE_IN_WORDS ];
   if (((UintPtr)val & 3) != 0) {
     memcpy(tempData, val, len);
     val = tempData;
   }
 
-  if (insertATTRINFO(Interpreter::BranchCol(c, 0, 0, false)) == -1)
+  if (insertATTRINFO(Interpreter::BranchCol(c, 0, 0)) == -1)
     DBUG_RETURN(-1);
   
   if (insertBranch(Label) == -1)
@@ -1070,7 +1157,8 @@ NdbOperation::branch_col(Uint32 type,
     DBUG_RETURN(-1);
   
   Uint32 len2 = Interpreter::mod4(len);
-  if(len2 == len){
+  if((len2 == len) &&
+     (lastWordMask == (Uint32)~0)){
     insertATTRINFOloop((Uint32*)val, len2 >> 2);
   } else {
     len2 -= 4;
@@ -1080,7 +1168,7 @@ NdbOperation::branch_col(Uint32 type,
       char* p = (char*)&tmp;
       p[i] = ((char*)val)[len2+i];
     }
-    insertATTRINFO(tmp);
+    insertATTRINFO(tmp & lastWordMask);
   }
   
   theErrorLine++;
@@ -1092,7 +1180,7 @@ NdbOperation::branch_col_eq(Uint32 ColId, const void * val, Uint32 len,
 			    bool nopad, Uint32 Label){
   INT_DEBUG(("branch_col_eq %u %.*s(%u,%d) -> %u", ColId, len, (char*) val, len,
              nopad, Label));
-  return branch_col(Interpreter::EQ, ColId, val, len, nopad, Label);
+  return branch_col(Interpreter::EQ, ColId, val, len, Label);
 }
 
 int
@@ -1100,28 +1188,28 @@ NdbOperation::branch_col_ne(Uint32 ColId, const void * val, Uint32 len,
 			    bool nopad, Uint32 Label){
   INT_DEBUG(("branch_col_ne %u %.*s(%u,%d) -> %u", ColId, len, (char*) val, len,
              nopad, Label));
-  return branch_col(Interpreter::NE, ColId, val, len, nopad, Label);
+  return branch_col(Interpreter::NE, ColId, val, len, Label);
 }
 int
 NdbOperation::branch_col_lt(Uint32 ColId, const void * val, Uint32 len, 
 			    bool nopad, Uint32 Label){
   INT_DEBUG(("branch_col_lt %u %.*s(%u,%d) -> %u", ColId, len, (char*) val, len,
              nopad, Label));
-  return branch_col(Interpreter::LT, ColId, val, len, nopad, Label);
+  return branch_col(Interpreter::LT, ColId, val, len, Label);
 }
 int
 NdbOperation::branch_col_le(Uint32 ColId, const void * val, Uint32 len, 
 			    bool nopad, Uint32 Label){
   INT_DEBUG(("branch_col_le %u %.*s(%u,%d) -> %u", ColId, len, (char*) val, len,
              nopad, Label));
-  return branch_col(Interpreter::LE, ColId, val, len, nopad, Label);
+  return branch_col(Interpreter::LE, ColId, val, len, Label);
 }
 int
 NdbOperation::branch_col_gt(Uint32 ColId, const void * val, Uint32 len, 
 			    bool nopad, Uint32 Label){
   INT_DEBUG(("branch_col_gt %u %.*s(%u,%d) -> %u", ColId, len, (char*) val, len,
              nopad, Label));
-  return branch_col(Interpreter::GT, ColId, val, len, nopad, Label);
+  return branch_col(Interpreter::GT, ColId, val, len, Label);
 }
 
 int
@@ -1129,7 +1217,7 @@ NdbOperation::branch_col_ge(Uint32 ColId, const void * val, Uint32 len,
 			    bool nopad, Uint32 Label){
   INT_DEBUG(("branch_col_ge %u %.*s(%u,%d) -> %u", ColId, len, (char*) val, len,
              nopad, Label));
-  return branch_col(Interpreter::GE, ColId, val, len, nopad, Label);
+  return branch_col(Interpreter::GE, ColId, val, len, Label);
 }
 
 int
@@ -1137,7 +1225,7 @@ NdbOperation::branch_col_like(Uint32 ColId, const void * val, Uint32 len,
 			      bool nopad, Uint32 Label){
   INT_DEBUG(("branch_col_like %u %.*s(%u,%d) -> %u", ColId, len, (char*) val, len,
              nopad, Label));
-  return branch_col(Interpreter::LIKE, ColId, val, len, nopad, Label);
+  return branch_col(Interpreter::LIKE, ColId, val, len, Label);
 }
 
 int
@@ -1145,7 +1233,39 @@ NdbOperation::branch_col_notlike(Uint32 ColId, const void * val, Uint32 len,
 				 bool nopad, Uint32 Label){
   INT_DEBUG(("branch_col_notlike %u %.*s(%u,%d) -> %u", ColId, len, (char*) val, len,
              nopad, Label));
-  return branch_col(Interpreter::NOT_LIKE, ColId, val, len, nopad, Label);
+  return branch_col(Interpreter::NOT_LIKE, ColId, val, len, Label);
+}
+
+int
+NdbOperation::branch_col_and_mask_eq_mask(Uint32 ColId, const void * mask, 
+                                          Uint32 len, bool nopad, Uint32 Label){
+  INT_DEBUG(("branch_col_and_mask_eq_mask %u %.*s(%u,%d) -> %u", ColId, len, (char*) mask, len,
+             nopad, Label));
+  return branch_col(Interpreter::AND_EQ_MASK, ColId, mask, len, Label);
+}
+
+int
+NdbOperation::branch_col_and_mask_ne_mask(Uint32 ColId, const void * mask, 
+                                          Uint32 len, bool nopad, Uint32 Label){
+  INT_DEBUG(("branch_col_and_mask_ne_mask %u %.*s(%u,%d) -> %u", ColId, len, (char*) mask, len,
+             nopad, Label));
+  return branch_col(Interpreter::AND_NE_MASK, ColId, mask, len, Label);
+}
+
+int
+NdbOperation::branch_col_and_mask_eq_zero(Uint32 ColId, const void * mask, 
+                                          Uint32 len, bool nopad, Uint32 Label){
+  INT_DEBUG(("branch_col_and_mask_eq_zero %u %.*s(%u,%d) -> %u", ColId, len, (char*) mask, len,
+             nopad, Label));
+  return branch_col(Interpreter::AND_EQ_ZERO, ColId, mask, len, Label);
+}
+
+int
+NdbOperation::branch_col_and_mask_ne_zero(Uint32 ColId, const void * mask, 
+                                          Uint32 len, bool nopad, Uint32 Label){
+  INT_DEBUG(("branch_col_and_mask_ne_zero %u %.*s(%u,%d) -> %u", ColId, len, (char*) mask, len,
+             nopad, Label));
+  return branch_col(Interpreter::AND_NE_ZERO, ColId, mask, len, Label);
 }
 
 int
