@@ -1,19 +1,26 @@
-/* Copyright (c) 2003-2007 MySQL AB
-
+/*
+   Copyright (c) 2003, 2015, Oracle and/or its affiliates. All rights reserved.
 
    This program is free software; you can redistribute it and/or modify
-   it under the terms of the GNU General Public License as published by
-   the Free Software Foundation; version 2 of the License.
+   it under the terms of the GNU General Public License, version 2.0,
+   as published by the Free Software Foundation.
+
+   This program is also distributed with certain software (including
+   but not limited to OpenSSL) that is licensed under separate terms,
+   as designated in a particular file or component or in included license
+   documentation.  The authors of MySQL hereby grant you an additional
+   permission to link the program and your derivative works with the
+   separately licensed software that they have included with MySQL.
 
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-   GNU General Public License for more details.
+   GNU General Public License, version 2.0, for more details.
 
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
-   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA */
-
+   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA
+*/
 
 #ifndef HUGO_TRANSACTIONS_HPP
 #define HUGO_TRANSACTIONS_HPP
@@ -141,6 +148,8 @@ public:
 			 int batchsize = 1);
 
   void setRetryMax(int retryMax = 100) { m_retryMax = retryMax; }
+  // XXX only for scanUpdateRecords
+  bool getRetryMaxReached() const { return m_retryMaxReached; }
   
   Uint64 m_latest_gci;
 
@@ -156,6 +165,7 @@ protected:
   NDBT_ResultRow row;
   int m_defaultScanUpdateMethod;
   int m_retryMax;
+  bool m_retryMaxReached;
 
   NDBT_Stats* m_stats_latency;
 
